#define _GNU_SOURCE
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "missing.h"
#include "fanotify.h"

int readlink_malloc(const char *p, char **r) {
        size_t l = 100;

        for (;;) {
                char *c;
                ssize_t n;

                if (!(c = malloc(sizeof(char) * l)))
                        return -ENOMEM;

                if ((n = readlink(p, c, l-1)) < 0) {
                        int ret = -errno;
                        free(c);
                        return ret;
                }

                if ((size_t) n < l-1) {
                        c[n] = 0;
                        *r = c;
                        return 0;
                }

                free(c);
                l *= 2;
        }
}

int main(int argc, char *argv[]) {
	int fanotify_fd;
	char *root="/usr";

	fprintf(stderr, "Testing fanotify...\n");

	fprintf(stderr, "fanotify_fd = fanotify_init(FAN_CLOEXEC|FAN_NONBLOCK<0x%x>,\n"
			"\tO_RDONLY|O_LARGEFILE|O_CLOEXEC|O_NOATIME<0x%x>)\n",
			FAN_CLOEXEC,
			O_RDONLY|O_LARGEFILE|O_CLOEXEC|O_NOATIME);
	if ((fanotify_fd = fanotify_init(FAN_CLOEXEC, O_RDONLY|O_LARGEFILE|O_CLOEXEC|O_NOATIME)) < 0)  {
                fprintf(stderr, "Failed to create fanotify object: (%s)", strerror(errno));
		exit(1);
        }

	fprintf(stderr, "fanotify_mark(fanotify_fd<%d>,\n"
			"\tFAN_MARK_ADD|FAN_MARK_MOUNT<flags:0x%d>\n"
			"\tFAN_OPEN<mask:0x%x>\n"
			"\tAT_FDCWD<dfd:%d>\n"
			"\tpathname = %s<%p>\n",
			fanotify_fd, FAN_MARK_ADD|FAN_MARK_MOUNT, FAN_OPEN, AT_FDCWD, root, root);
        if (fanotify_mark(fanotify_fd, FAN_MARK_ADD|FAN_MARK_MOUNT, FAN_OPEN, AT_FDCWD, root) < 0) {
		fprintf(stderr, "Failed to mark '%s' (%s)", root, strerror(errno));
		exit(1);
        }

	for (;;) {
                union {
                        struct fanotify_event_metadata metadata;
                        char buffer[4096];
                } data;
                ssize_t n;
                struct fanotify_event_metadata *m;

		fprintf(stderr, "fanotify starting monitor...\n");
                if ((n = read(fanotify_fd, &data, sizeof(data))) < 0) {

                        if (errno == EINTR || errno == EAGAIN)
                                continue;

                        fprintf(stderr, "Failed to read event(empty), quit...\n");
			exit(1);
                }
		fprintf(stderr, "read %d events...\n", n);

                for (m = &data.metadata; FAN_EVENT_OK(m, n); m = FAN_EVENT_NEXT(m, n)) {
                        char fn[4096] = {0,};
			char *p;
                        int k;

                        if (m->fd < 0) continue;
                	snprintf(fn, sizeof(fn), "/proc/self/fd/%i", m->fd);
			readlink_malloc(fn, &p);
			fprintf(stderr, "\tAccess '%s'\n", p);
			free(p);
	
        		for (;;) {                      
                		int r;                          
                                 
                		if ((r = close(m->fd)) >= 0)
                        		break;

                		if (errno != EINTR)
					break;
        		}

                }
        }



        return 0;
}
